"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.export.tfsm_layer import TFSMLayer as TFSMLayer
from keras.src.layers import deserialize as deserialize
from keras.src.layers import serialize as serialize
from keras.src.layers.activations.activation import Activation as Activation
from keras.src.layers.activations.elu import ELU as ELU
from keras.src.layers.activations.leaky_relu import LeakyReLU as LeakyReLU
from keras.src.layers.activations.prelu import PReLU as PReLU
from keras.src.layers.activations.relu import ReLU as ReLU
from keras.src.layers.activations.softmax import Softmax as Softmax
from keras.src.layers.attention.additive_attention import (
    AdditiveAttention as AdditiveAttention,
)
from keras.src.layers.attention.attention import Attention as Attention
from keras.src.layers.attention.grouped_query_attention import (
    GroupedQueryAttention as GroupQueryAttention,
)
from keras.src.layers.attention.multi_head_attention import (
    MultiHeadAttention as MultiHeadAttention,
)
from keras.src.layers.convolutional.conv1d import Conv1D as Conv1D
from keras.src.layers.convolutional.conv1d import Conv1D as Convolution1D
from keras.src.layers.convolutional.conv1d_transpose import (
    Conv1DTranspose as Conv1DTranspose,
)
from keras.src.layers.convolutional.conv1d_transpose import (
    Conv1DTranspose as Convolution1DTranspose,
)
from keras.src.layers.convolutional.conv2d import Conv2D as Conv2D
from keras.src.layers.convolutional.conv2d import Conv2D as Convolution2D
from keras.src.layers.convolutional.conv2d_transpose import (
    Conv2DTranspose as Conv2DTranspose,
)
from keras.src.layers.convolutional.conv2d_transpose import (
    Conv2DTranspose as Convolution2DTranspose,
)
from keras.src.layers.convolutional.conv3d import Conv3D as Conv3D
from keras.src.layers.convolutional.conv3d import Conv3D as Convolution3D
from keras.src.layers.convolutional.conv3d_transpose import (
    Conv3DTranspose as Conv3DTranspose,
)
from keras.src.layers.convolutional.conv3d_transpose import (
    Conv3DTranspose as Convolution3DTranspose,
)
from keras.src.layers.convolutional.depthwise_conv1d import (
    DepthwiseConv1D as DepthwiseConv1D,
)
from keras.src.layers.convolutional.depthwise_conv2d import (
    DepthwiseConv2D as DepthwiseConv2D,
)
from keras.src.layers.convolutional.separable_conv1d import (
    SeparableConv1D as SeparableConv1D,
)
from keras.src.layers.convolutional.separable_conv1d import (
    SeparableConv1D as SeparableConvolution1D,
)
from keras.src.layers.convolutional.separable_conv2d import (
    SeparableConv2D as SeparableConv2D,
)
from keras.src.layers.convolutional.separable_conv2d import (
    SeparableConv2D as SeparableConvolution2D,
)
from keras.src.layers.core.dense import Dense as Dense
from keras.src.layers.core.einsum_dense import EinsumDense as EinsumDense
from keras.src.layers.core.embedding import Embedding as Embedding
from keras.src.layers.core.identity import Identity as Identity
from keras.src.layers.core.input_layer import Input as Input
from keras.src.layers.core.input_layer import InputLayer as InputLayer
from keras.src.layers.core.lambda_layer import Lambda as Lambda
from keras.src.layers.core.masking import Masking as Masking
from keras.src.layers.core.reversible_embedding import (
    ReversibleEmbedding as ReversibleEmbedding,
)
from keras.src.layers.core.wrapper import Wrapper as Wrapper
from keras.src.layers.input_spec import InputSpec as InputSpec
from keras.src.layers.layer import Layer as Layer
from keras.src.layers.merging.add import Add as Add
from keras.src.layers.merging.add import add as add
from keras.src.layers.merging.average import Average as Average
from keras.src.layers.merging.average import average as average
from keras.src.layers.merging.concatenate import Concatenate as Concatenate
from keras.src.layers.merging.concatenate import concatenate as concatenate
from keras.src.layers.merging.dot import Dot as Dot
from keras.src.layers.merging.dot import dot as dot
from keras.src.layers.merging.maximum import Maximum as Maximum
from keras.src.layers.merging.maximum import maximum as maximum
from keras.src.layers.merging.minimum import Minimum as Minimum
from keras.src.layers.merging.minimum import minimum as minimum
from keras.src.layers.merging.multiply import Multiply as Multiply
from keras.src.layers.merging.multiply import multiply as multiply
from keras.src.layers.merging.subtract import Subtract as Subtract
from keras.src.layers.merging.subtract import subtract as subtract
from keras.src.layers.normalization.batch_normalization import (
    BatchNormalization as BatchNormalization,
)
from keras.src.layers.normalization.group_normalization import (
    GroupNormalization as GroupNormalization,
)
from keras.src.layers.normalization.layer_normalization import (
    LayerNormalization as LayerNormalization,
)
from keras.src.layers.normalization.rms_normalization import (
    RMSNormalization as RMSNormalization,
)
from keras.src.layers.normalization.spectral_normalization import (
    SpectralNormalization as SpectralNormalization,
)
from keras.src.layers.normalization.unit_normalization import (
    UnitNormalization as UnitNormalization,
)
from keras.src.layers.pooling.adaptive_average_pooling1d import (
    AdaptiveAveragePooling1D as AdaptiveAveragePooling1D,
)
from keras.src.layers.pooling.adaptive_average_pooling2d import (
    AdaptiveAveragePooling2D as AdaptiveAveragePooling2D,
)
from keras.src.layers.pooling.adaptive_average_pooling3d import (
    AdaptiveAveragePooling3D as AdaptiveAveragePooling3D,
)
from keras.src.layers.pooling.adaptive_max_pooling1d import (
    AdaptiveMaxPooling1D as AdaptiveMaxPooling1D,
)
from keras.src.layers.pooling.adaptive_max_pooling2d import (
    AdaptiveMaxPooling2D as AdaptiveMaxPooling2D,
)
from keras.src.layers.pooling.adaptive_max_pooling3d import (
    AdaptiveMaxPooling3D as AdaptiveMaxPooling3D,
)
from keras.src.layers.pooling.average_pooling1d import (
    AveragePooling1D as AveragePooling1D,
)
from keras.src.layers.pooling.average_pooling1d import (
    AveragePooling1D as AvgPool1D,
)
from keras.src.layers.pooling.average_pooling2d import (
    AveragePooling2D as AveragePooling2D,
)
from keras.src.layers.pooling.average_pooling2d import (
    AveragePooling2D as AvgPool2D,
)
from keras.src.layers.pooling.average_pooling3d import (
    AveragePooling3D as AveragePooling3D,
)
from keras.src.layers.pooling.average_pooling3d import (
    AveragePooling3D as AvgPool3D,
)
from keras.src.layers.pooling.global_average_pooling1d import (
    GlobalAveragePooling1D as GlobalAveragePooling1D,
)
from keras.src.layers.pooling.global_average_pooling1d import (
    GlobalAveragePooling1D as GlobalAvgPool1D,
)
from keras.src.layers.pooling.global_average_pooling2d import (
    GlobalAveragePooling2D as GlobalAveragePooling2D,
)
from keras.src.layers.pooling.global_average_pooling2d import (
    GlobalAveragePooling2D as GlobalAvgPool2D,
)
from keras.src.layers.pooling.global_average_pooling3d import (
    GlobalAveragePooling3D as GlobalAveragePooling3D,
)
from keras.src.layers.pooling.global_average_pooling3d import (
    GlobalAveragePooling3D as GlobalAvgPool3D,
)
from keras.src.layers.pooling.global_max_pooling1d import (
    GlobalMaxPooling1D as GlobalMaxPool1D,
)
from keras.src.layers.pooling.global_max_pooling1d import (
    GlobalMaxPooling1D as GlobalMaxPooling1D,
)
from keras.src.layers.pooling.global_max_pooling2d import (
    GlobalMaxPooling2D as GlobalMaxPool2D,
)
from keras.src.layers.pooling.global_max_pooling2d import (
    GlobalMaxPooling2D as GlobalMaxPooling2D,
)
from keras.src.layers.pooling.global_max_pooling3d import (
    GlobalMaxPooling3D as GlobalMaxPool3D,
)
from keras.src.layers.pooling.global_max_pooling3d import (
    GlobalMaxPooling3D as GlobalMaxPooling3D,
)
from keras.src.layers.pooling.max_pooling1d import MaxPooling1D as MaxPool1D
from keras.src.layers.pooling.max_pooling1d import MaxPooling1D as MaxPooling1D
from keras.src.layers.pooling.max_pooling2d import MaxPooling2D as MaxPool2D
from keras.src.layers.pooling.max_pooling2d import MaxPooling2D as MaxPooling2D
from keras.src.layers.pooling.max_pooling3d import MaxPooling3D as MaxPool3D
from keras.src.layers.pooling.max_pooling3d import MaxPooling3D as MaxPooling3D
from keras.src.layers.preprocessing.category_encoding import (
    CategoryEncoding as CategoryEncoding,
)
from keras.src.layers.preprocessing.discretization import (
    Discretization as Discretization,
)
from keras.src.layers.preprocessing.hashed_crossing import (
    HashedCrossing as HashedCrossing,
)
from keras.src.layers.preprocessing.hashing import Hashing as Hashing
from keras.src.layers.preprocessing.image_preprocessing.aug_mix import (
    AugMix as AugMix,
)
from keras.src.layers.preprocessing.image_preprocessing.auto_contrast import (
    AutoContrast as AutoContrast,
)
from keras.src.layers.preprocessing.image_preprocessing.center_crop import (
    CenterCrop as CenterCrop,
)
from keras.src.layers.preprocessing.image_preprocessing.clahe import (
    ContrastLimitedAdaptiveHistogramEqualization as ContrastLimitedAdaptiveHistogramEqualization,
)
from keras.src.layers.preprocessing.image_preprocessing.cut_mix import (
    CutMix as CutMix,
)
from keras.src.layers.preprocessing.image_preprocessing.equalization import (
    Equalization as Equalization,
)
from keras.src.layers.preprocessing.image_preprocessing.max_num_bounding_box import (
    MaxNumBoundingBoxes as MaxNumBoundingBoxes,
)
from keras.src.layers.preprocessing.image_preprocessing.mix_up import (
    MixUp as MixUp,
)
from keras.src.layers.preprocessing.image_preprocessing.rand_augment import (
    RandAugment as RandAugment,
)
from keras.src.layers.preprocessing.image_preprocessing.random_brightness import (
    RandomBrightness as RandomBrightness,
)
from keras.src.layers.preprocessing.image_preprocessing.random_color_degeneration import (
    RandomColorDegeneration as RandomColorDegeneration,
)
from keras.src.layers.preprocessing.image_preprocessing.random_color_jitter import (
    RandomColorJitter as RandomColorJitter,
)
from keras.src.layers.preprocessing.image_preprocessing.random_contrast import (
    RandomContrast as RandomContrast,
)
from keras.src.layers.preprocessing.image_preprocessing.random_crop import (
    RandomCrop as RandomCrop,
)
from keras.src.layers.preprocessing.image_preprocessing.random_elastic_transform import (
    RandomElasticTransform as RandomElasticTransform,
)
from keras.src.layers.preprocessing.image_preprocessing.random_erasing import (
    RandomErasing as RandomErasing,
)
from keras.src.layers.preprocessing.image_preprocessing.random_flip import (
    RandomFlip as RandomFlip,
)
from keras.src.layers.preprocessing.image_preprocessing.random_gaussian_blur import (
    RandomGaussianBlur as RandomGaussianBlur,
)
from keras.src.layers.preprocessing.image_preprocessing.random_grayscale import (
    RandomGrayscale as RandomGrayscale,
)
from keras.src.layers.preprocessing.image_preprocessing.random_hue import (
    RandomHue as RandomHue,
)
from keras.src.layers.preprocessing.image_preprocessing.random_invert import (
    RandomInvert as RandomInvert,
)
from keras.src.layers.preprocessing.image_preprocessing.random_perspective import (
    RandomPerspective as RandomPerspective,
)
from keras.src.layers.preprocessing.image_preprocessing.random_posterization import (
    RandomPosterization as RandomPosterization,
)
from keras.src.layers.preprocessing.image_preprocessing.random_rotation import (
    RandomRotation as RandomRotation,
)
from keras.src.layers.preprocessing.image_preprocessing.random_saturation import (
    RandomSaturation as RandomSaturation,
)
from keras.src.layers.preprocessing.image_preprocessing.random_sharpness import (
    RandomSharpness as RandomSharpness,
)
from keras.src.layers.preprocessing.image_preprocessing.random_shear import (
    RandomShear as RandomShear,
)
from keras.src.layers.preprocessing.image_preprocessing.random_translation import (
    RandomTranslation as RandomTranslation,
)
from keras.src.layers.preprocessing.image_preprocessing.random_zoom import (
    RandomZoom as RandomZoom,
)
from keras.src.layers.preprocessing.image_preprocessing.resizing import (
    Resizing as Resizing,
)
from keras.src.layers.preprocessing.image_preprocessing.solarization import (
    Solarization as Solarization,
)
from keras.src.layers.preprocessing.integer_lookup import (
    IntegerLookup as IntegerLookup,
)
from keras.src.layers.preprocessing.mel_spectrogram import (
    MelSpectrogram as MelSpectrogram,
)
from keras.src.layers.preprocessing.normalization import (
    Normalization as Normalization,
)
from keras.src.layers.preprocessing.pipeline import Pipeline as Pipeline
from keras.src.layers.preprocessing.rescaling import Rescaling as Rescaling
from keras.src.layers.preprocessing.stft_spectrogram import (
    STFTSpectrogram as STFTSpectrogram,
)
from keras.src.layers.preprocessing.string_lookup import (
    StringLookup as StringLookup,
)
from keras.src.layers.preprocessing.text_vectorization import (
    TextVectorization as TextVectorization,
)
from keras.src.layers.regularization.activity_regularization import (
    ActivityRegularization as ActivityRegularization,
)
from keras.src.layers.regularization.dropout import Dropout as Dropout
from keras.src.layers.regularization.gaussian_dropout import (
    GaussianDropout as GaussianDropout,
)
from keras.src.layers.regularization.gaussian_noise import (
    GaussianNoise as GaussianNoise,
)
from keras.src.layers.regularization.spatial_dropout import (
    SpatialDropout1D as SpatialDropout1D,
)
from keras.src.layers.regularization.spatial_dropout import (
    SpatialDropout2D as SpatialDropout2D,
)
from keras.src.layers.regularization.spatial_dropout import (
    SpatialDropout3D as SpatialDropout3D,
)
from keras.src.layers.reshaping.cropping1d import Cropping1D as Cropping1D
from keras.src.layers.reshaping.cropping2d import Cropping2D as Cropping2D
from keras.src.layers.reshaping.cropping3d import Cropping3D as Cropping3D
from keras.src.layers.reshaping.flatten import Flatten as Flatten
from keras.src.layers.reshaping.permute import Permute as Permute
from keras.src.layers.reshaping.repeat_vector import (
    RepeatVector as RepeatVector,
)
from keras.src.layers.reshaping.reshape import Reshape as Reshape
from keras.src.layers.reshaping.up_sampling1d import (
    UpSampling1D as UpSampling1D,
)
from keras.src.layers.reshaping.up_sampling2d import (
    UpSampling2D as UpSampling2D,
)
from keras.src.layers.reshaping.up_sampling3d import (
    UpSampling3D as UpSampling3D,
)
from keras.src.layers.reshaping.zero_padding1d import (
    ZeroPadding1D as ZeroPadding1D,
)
from keras.src.layers.reshaping.zero_padding2d import (
    ZeroPadding2D as ZeroPadding2D,
)
from keras.src.layers.reshaping.zero_padding3d import (
    ZeroPadding3D as ZeroPadding3D,
)
from keras.src.layers.rnn.bidirectional import Bidirectional as Bidirectional
from keras.src.layers.rnn.conv_lstm1d import ConvLSTM1D as ConvLSTM1D
from keras.src.layers.rnn.conv_lstm2d import ConvLSTM2D as ConvLSTM2D
from keras.src.layers.rnn.conv_lstm3d import ConvLSTM3D as ConvLSTM3D
from keras.src.layers.rnn.gru import GRU as GRU
from keras.src.layers.rnn.gru import GRUCell as GRUCell
from keras.src.layers.rnn.lstm import LSTM as LSTM
from keras.src.layers.rnn.lstm import LSTMCell as LSTMCell
from keras.src.layers.rnn.rnn import RNN as RNN
from keras.src.layers.rnn.simple_rnn import SimpleRNN as SimpleRNN
from keras.src.layers.rnn.simple_rnn import SimpleRNNCell as SimpleRNNCell
from keras.src.layers.rnn.stacked_rnn_cells import (
    StackedRNNCells as StackedRNNCells,
)
from keras.src.layers.rnn.time_distributed import (
    TimeDistributed as TimeDistributed,
)
from keras.src.legacy.layers import AlphaDropout as AlphaDropout
from keras.src.legacy.layers import RandomHeight as RandomHeight
from keras.src.legacy.layers import RandomWidth as RandomWidth
from keras.src.legacy.layers import ThresholdedReLU as ThresholdedReLU
from keras.src.utils.jax_layer import FlaxLayer as FlaxLayer
from keras.src.utils.jax_layer import JaxLayer as JaxLayer
from keras.src.utils.torch_utils import TorchModuleWrapper as TorchModuleWrapper



"""A class for JAX specific optimizer logic.

Its purpose is to route around statelessness
requirements in cond ops used for EMA handling
and gradient accumulation handling. We do this
by skipping conditionals entirely.
"""

import jax
from jax import numpy as jnp

from keras.src.optimizers import base_optimizer


class JaxOptimizer(base_optimizer.BaseOptimizer):
    def _backend_apply_gradients(self, grads, trainable_variables):
        if self.gradient_accumulation_steps:
            is_update_step = (
                self._iterations + 1
            ) % self.gradient_accumulation_steps == 0
            steps = self.gradient_accumulation_steps

            current_trainable_vars_value = [
                v.value for v in trainable_variables
            ]
            current_optimizer_vars_value = [v.value for v in self.variables]

            # `trainable_variables` might have been filtered in previous
            # processing steps, so we need to ensure the correct mapping between
            # `self._accumulated_gradients` and `trainable_variables`
            acc_grads = [
                self._accumulated_gradients[self._get_variable_index(v)]
                for v in trainable_variables
            ]

            new_g_accs = jax.lax.cond(
                is_update_step,
                lambda: [jnp.zeros(g.shape, dtype=g.dtype) for g in acc_grads],
                lambda: [g + acc_g.value for g, acc_g in zip(grads, acc_grads)],
            )

            grads = jax.lax.cond(
                is_update_step,
                lambda: [
                    (g + acc_g.value) / steps
                    for g, acc_g in zip(grads, acc_grads)
                ],
                lambda: list(grads),
            )

            # Apply clipping and weight decay.
            grads = self._clip_gradients(grads)
            self._apply_weight_decay(trainable_variables)

            self._backend_update_step(
                grads, trainable_variables, self.learning_rate
            )
            new_trainable_vars = jax.lax.cond(
                is_update_step,
                lambda: [v.value for v in trainable_variables],
                lambda: current_trainable_vars_value,
            )
            new_opt_vars = jax.lax.cond(
                is_update_step,
                lambda: [v.value for v in self.variables],
                lambda: current_optimizer_vars_value,
            )

            for value, v in zip(new_trainable_vars, trainable_variables):
                v.assign(value)

            for value, v in zip(new_opt_vars, self.variables):
                v.assign(value)

            for n_g_acc, g_acc in zip(new_g_accs, acc_grads):
                g_acc.assign(n_g_acc)

        else:
            # Apply clipping and weight decay.
            grads = self._clip_gradients(grads)
            self._apply_weight_decay(trainable_variables)

            self._backend_update_step(
                grads, trainable_variables, self.learning_rate
            )

        if self.use_ema:
            self._update_model_variables_moving_average(
                self._trainable_variables
            )
            if self.ema_overwrite_frequency is not None:
                should_overwrite_model_vars = (
                    self.iterations + 1
                ) % self.ema_overwrite_frequency == 0
                should_overwrite_model_vars_int = (
                    should_overwrite_model_vars.astype("int32")
                )
                should_not_overwrite_model_vars_int = jnp.logical_not(
                    should_overwrite_model_vars
                ).astype("int32")
                current_trainable_vars_value = [
                    v.value for v in self._trainable_variables
                ]
                for var, average_var in zip(
                    self._trainable_variables,
                    self._model_variables_moving_average,
                ):
                    var.assign(
                        average_var * should_overwrite_model_vars_int
                        + var.value * should_not_overwrite_model_vars_int
                    )
